/*
 * This file is part of Codespeak.
 *
 * Codespeak is free software: you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, 
 * or (at your option) any later version.
 *
 * Codespeak is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with Codespeak. 
 * If not, see <https://www.gnu.org/licenses/>.
 * 
 *
 * 13.10.2022 created.
 * gcc -static -I../local/include -L../local/lib -Wall -g codk_expression.c -lgrapheme
 *
 */

#include <grapheme.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/* 	expr to list of lists, as identified by language	 */

/* Keywords of length 1 split words: */
#define CODK_NKC 2
char *KEYCHAR[CODK_NKC] = { "(", ")" };  /* later: "... */

signed char
keychar(char *s, size_t l) {
	for (signed char i=0; i<CODK_NKC; ++i) {
		if (strncmp(KEYCHAR[i], s, l) == 0) {
			return i;
		}
	}
	return -1;
}

typedef struct _list {  
	struct _list *prev;
	struct _list *next;
	char *val;
} list;

static list *
new_list(void) {
	list *l = malloc(sizeof(list));
	l->prev = l->next = NULL;
	l->val = NULL;
	return l;
}
static void
free_list_val(list *l) {
	free(l->val);
	if (l->next != NULL) {
		free_list_val(l->next);
		free(l);
	}
}
static void
free_list(list *l) {
	if (l == NULL) {
		return;
	}
	free_list_val(l);
}
static list *
root(list *l) {
	if (l == NULL) {
		return l;
	}
	while (l->prev) {
		l = l->prev;
	}
	return l;
}
static void
show_list(list *l) {
	if (l == NULL) {
		return;
	}
	l = root(l);
	while (l) {
		printf("%s ", l->val);
		l = l->next;
	}
	printf("\n");
}
static list *
list_add(list *l, char *s, size_t sl) {
	list *x = new_list();
	char *y = malloc((sl+1)*sizeof(char));
	memcpy(y, s, sl);
	y[sl] = '\0';
	x->val = y;
	if (l) {
		x->prev = l;
		l->next = x;
	}
	return x;
}
#define CODK_WSZ 127
static list * 
parse_words(char *s) {
	list *l = NULL;
	size_t boff = 0;
	char buf[CODK_WSZ+1];
	buf[0] = '\0';
	size_t read;
	for (size_t off = 0; s[off] != '\0'; off += read) {
		read = grapheme_next_character_break_utf8(s+off, SIZE_MAX);
		printf("<%.*s> ", (int)read, s+off);
		if (keychar(s+off, read) >= 0) {
			if (boff) {
				l = list_add(l, buf, boff);
				boff = 0;
			}
			l = list_add(l, s+off, read);
			continue;
		}
		if (isspace((int)*(s+off))) {
			if (boff) {
				l = list_add(l, buf, boff);
				boff = 0;
			}
			continue;
		}
		if (boff+read > CODK_WSZ) {
			printf("error: word too big (%luB)!\n", boff+read);
			free_list(l);
			return NULL;
		}
		memcpy(buf+boff, s+off, read*sizeof(char));
		boff += read;
		buf[boff] = '\0';
	}
	if (boff) {
		l = list_add(l, buf, boff);
	}
	printf("\n");
	return l;
}

static void 
parse_expr(char *s_arg) {
	char *s = strdup(s_arg);
	/* split string into lists, for next step */
	list *w = parse_words(s);
	show_list(w);
	free_list(w);
	/* cleanup */
	free(s);
	s = NULL;
}

/* 	expr to list of lists, as identified by language	 */

static void
usage(const char *exe) {
	printf("usage: %s expression\n", exe);
}

int
main(int argc, char **argv) {
	if (argc < 2) {
		usage(argv[0]);
		return 1;
	}
	char *s = argv[1];
	printf("input: <%s>\n", s);
	parse_expr(s);
	return 0;
}

